package de.uni_passau.fim.seibt.gitwrapper.repo;

import java.io.File;
import java.nio.file.Path;
import java.util.List;
import java.util.Objects;
import java.util.Optional;

import org.apache.commons.io.FileUtils;

/**
 * A git repository providing methods for accessing its state.
 */
public abstract class Repository {

    /**
     * Performs a checkout of the given {@link Reference}.
     *
     * @param ref
     *         the {@link Reference} to checkout
     * @return whether the checkout was successful
     * @see <a href=https://git-scm.com/docs/git-checkout>git checkout</a>
     */
    public abstract boolean checkout(Reference ref);

    /**
     * Performs a checkout of the given {@link Reference}. All changes since last commit will be discarded.
     *
     * @param ref
     *         the {@link Reference} to checkout
     * @return whether the checkout was successful
     * @see <a href=https://git-scm.com/docs/git-checkout>git checkout</a>
     */
    public abstract boolean forceCheckout(Reference ref);

    /**
     * Performs a fetch in this {@link Repository}.
     *
     * @return whether the fetch was successful
     * @see <a href=https://git-scm.com/docs/git-fetch>git fetch</a>
     */
    public abstract boolean fetch();

    /**
     * Optionally returns a list of all merge commits in this repository. An empty {@link Optional} will be returned
     * if there is an error retrieving the merge commits.
     *
     * @return the list of merge commits
     */
    public abstract Optional<List<Commit>> getMergeCommits();

    /**
     * Returns the {@link Commit} currently pointed at by <code>HEAD</code>.
     *
     * @return the <code>HEAD</code> commit or an empty {@link Optional} if there was an error
     */
    public abstract Optional<Commit> getCurrentHEAD();

    /**
     * Returns a {@link Commit} for the given ID. The caller must ensure that the ID is a full SHA1 hash of a
     * commit that exists in this repository.
     *
     * @param id
     *         the ID for the {@link Commit}
     * @return the {@link Commit}
     *
     * @see #getCommit(String)
     */
    protected abstract Commit getCommitUnchecked(String id);

    /**
     * Optionally returns a {@link Commit} for the given ID. If the given ID does not designate a commit that exists in
     * this {@link Repository}, an empty {@link Optional} will be returned. The ID will be resolved to a full SHA1 hash.
     *
     * @param id
     *         the ID of the commit
     * @return the {@link Commit} or an empty {@link Optional} if the ID is invalid or an exception occurs
     */
    public abstract Optional<Commit> getCommit(String id);

    /**
     * Optionally returns a {@link Branch} for the given name. If the given name does not designate a branch that
     * exists in this {@link Repository}, an empty {@link Optional} will be returned.
     *
     * @param name
     *         the name of the branch
     * @return the {@link Branch} or an empty {@link Optional} if the name is invalid or an exception occurs
     */
    public abstract Optional<Branch> getBranch(String name);

    /**
     * Resolves the given <code>id</code> to the full SHA1 hash.
     *
     * @param id
     *         the <code>id</code> to transform
     * @return the full SHA1 hash or an empty {@link Optional} if an exception occurs or {@code id} can not be converted
     *         to its associated git hash
     */
    protected abstract Optional<String> toHash(String id);

    /**
     * Copies this {@link Repository} to the given <code>destination</code> directory.
     *
     * @param destination
     *         the destination directory
     * @return a copy of this {@link Repository} at the new location
     * @see FileUtils#copyDirectory(File, File)
     */
    public abstract Optional<Repository> copy(File destination);

    /**
     * Parses the output of 'git blame' for the given {@code file} and returns a list of {@link BlameLine BlameLines}
     * containing the resulting information about every line of the file. If there is an exception parsing or obtaining
     * the 'git blame' output, an empty {@link Optional} will be returned.
     *
     * @param file
     *         the path to the file to analyze (relative to the git root or absolute)
     * @return optionally the list of {@link BlameLine BlameLines}
     */
    public abstract Optional<List<BlameLine>> blameFile(Path file);

    /**
     * Parses the output of 'git blame' for an unmerged file and extracts the merge conflicts that occurred in the file.
     * Every merge conflict is represented by a {@link MergeConflict} instance containing the {@link BlameLine BlameLines}
     * that represent the lines involved in the conflict. If there are no conflicts in the given file, an empty list
     * will be returned. If there is an exception parsing or obtaining the 'git blame' output, an empty {@link Optional}
     * will be returned.
     *
     * @param file
     *         the path to the unmerged file to analyze (relative to the git root or absolute)
     * @return optionally the list of {@link MergeConflict MergeConflicts}
     * @see #blameFile(Path)
     */
    public abstract Optional<List<MergeConflict>> blameUnmergedFile(Path file);

    /**
     * Parses and returns the current output of 'git status' for this {@link Repository}. This includes untracked files
     * but excludes ignored files. If there is an exception parsing or obtaining the status output, an empty
     * {@link Optional} will be returned.
     *
     * @return optionally the {@link Status} of this {@link Repository}
     */
    public abstract Optional<Status> getStatus();

    /**
     * Parses and returns the current output of 'git status' for this {@link Repository}.
     * If there is an exception parsing or obtaining the status output, an empty {@link Optional} will be returned.
     *
     * @param ignored
     *         whether to include ignored files in the {@link Status}
     * @param untracked
     *         whether to include untracked files in the {@link Status}
     * @return optionally the {@link Status} of this {@link Repository}
     */
    public abstract Optional<Status> getStatus(boolean ignored, boolean untracked);

    /**
     * Adds a remote to the this repository if it does not yet exist.
     *
     * @param name
     *         the name of the remote
     * @param forkURL
     *         the url of the remote
     * @return <code>true</code>, if adding was successful or the remote already exists, <code>false</code> otherwise
     */
    public abstract boolean addRemote(String name, String forkURL);

    /**
     * Creates a clean working directory at the currently checked out Reference
     *
     * @param retain
     *         patterns to be used as <a href=https://git-scm.com/docs/git-clean#git-clean--eltpatterngt>ignore rules</a>
     *
     * @return {@code true} if the cleanup was successful
     */
    public abstract boolean cleanup(String... retain);

    /**
     * Returns the {@link GitWrapper} used by this {@link Repository}.
     *
     * @return the {@link GitWrapper}
     */
    protected abstract GitWrapper getGit();

    /**
     * Returns the URL this {@link Repository} was cloned from.
     *
     * @return the URL
     */
    public abstract String getUrl();

    /**
     * Returns the directory containing this {@link Repository}.
     *
     * @return the repository directory
     */
    public abstract File getDir();

    /**
     * Returns the name of the directory this {@link Repository} is based in.
     *
     * @return the name of the {@link Repository}
     */
    public abstract String getName();

    @Override
    public final boolean equals(Object o) {
        if (this == o) {
            return true;
        }

        if (!(o instanceof Repository)) {
            return false;
        }

        Repository that = (Repository) o;
        return Objects.equals(getUrl(), that.getUrl())
                && Objects.equals(getDir().getAbsoluteFile(), that.getDir().getAbsoluteFile());
    }

    @Override
    public final int hashCode() {
        return Objects.hash(getUrl(), getDir().getAbsoluteFile());
    }
}
